open Common

let invalid_htl = Hashtbl.create 0

let f_installed i =
  match !Gl.pkgs.(i).status with
  | Current_version _  | Old_version _ | Local_version _ -> true
  | Unused  -> false

let f_not_installed i =
  match !Gl.pkgs.(i).status with
  | Unused  -> true
  | Current_version _ | Old_version _ | Local_version _ -> false

let f_updateable i =
  match !Gl.pkgs.(i).status with
  | Old_version _ -> true
  | Local_version _ | Unused | Current_version _ -> false



module Source = struct
  let filter_prefix str =
    let slen = String.length str in
    assert (slen > 0 );
    fun i ->
      let cstr = !Gl.pkgs.(i).name in
      let snlen = String.length cstr in
      if snlen < 1 || cstr.[0] <> str.[0] then
        false
      else if slen >= snlen then
        false
      else
        str = String.sub cstr 0 slen


  let combo_callback_filter_func' = ref Listview.always_true
  let combo_callback_last_modified = ref 0.0
  let combo_callback_real time () =
    if time = !combo_callback_last_modified then (
      let f = !combo_callback_filter_func' in
      Listview.s_view#set_filter_func f;
      combo_callback_last_modified :=  0.0;
    );
    false

  let combo_callback_last_text = ref ""
  let invalid_htl = Hashtbl.create 0
  let callback_now = ref false
  let callback () =
    let text = Util.remove_spaces_begin_end Toolbar.combo_source#entry#text in
    let time = Unix.gettimeofday () in
    combo_callback_last_modified := time ;
    if text <> !combo_callback_last_text then  (
      combo_callback_last_text := text;
      combo_callback_filter_func' := ( match text with
      | "" -> Listview.always_true
      | _ when text = Const.status_cat_updatable -> f_updateable
      | _ when text = Const.status_cat_installed -> f_installed
      | _ when text = Const.status_cat_not_installed -> f_not_installed
      | _ ->
        let ok =
          try
            fst ^$ List.find ( fun (_,b) -> b = text ) Const.categories
          with
          | Not_found -> Const.invalid_string
        in
        if ok != Const.invalid_string then
          filter_prefix ok
        else
          let text = String.lowercase text in
          let htl = try SHashtbl.find Gl.text_index_src text with Not_found -> invalid_htl in
          if htl == invalid_htl then
            Listview.always_false
          else (
            fun (i:int) -> Hashtbl.mem htl i )
      );
    );
    if !callback_now = false then
      ignore ( Glib.Timeout.add ~ms:200 ~callback:(combo_callback_real time) )
    else (
      callback_now := false;
      combo_callback_last_modified := 0. ;
      ignore ( combo_callback_real 0. () )
    )
end


module Binary = struct

  let combo_callback_last_modified = ref 0.0
  let combo_callback_filter_func' = ref Listview.always_true

  let combo_callback_last_text = ref ""
  let callback_now = ref false

  let l_status_cat_not_installed =  String.lowercase Const.status_cat_not_installed
  let l_status_cat_installed =  String.lowercase Const.status_cat_installed

  let combo_callback_real time () =
    if time = !combo_callback_last_modified then (
      let f = !combo_callback_filter_func' in
      Listview.b_view#set_filter_func f;
      combo_callback_last_modified :=  0.0;
    );
    false


  let callback () =
    let text = Util.remove_spaces_begin_end Toolbar.combo_bin#entry#text in
    let ltext =
      if text = Const.status_cat_installed then (
        l_status_cat_installed
      )
      else if text = Const.status_cat_not_installed then
        l_status_cat_not_installed
      else
        String.lowercase text
    in
    let time = Unix.gettimeofday () in
    combo_callback_last_modified := time ;
    if text <> !combo_callback_last_text then  (
      combo_callback_last_text := text;
      combo_callback_filter_func' := (
        if text = "" then
          Listview.always_true
        else if ltext == l_status_cat_not_installed then
          f_not_installed
        else if ltext == l_status_cat_installed then
          f_installed
        else
          let htl = try SHashtbl.find Gl.text_index_bin ltext with Not_found -> invalid_htl in
          if htl == invalid_htl then
            Listview.always_false
          else
            fun (i:int) -> Hashtbl.mem htl i
      )
    );
    if !callback_now = false then
      ignore ( Glib.Timeout.add ~ms:200 ~callback:(combo_callback_real time) )
    else (
      callback_now := false;
      combo_callback_last_modified := 0. ;
      ignore ( combo_callback_real 0. () )
    )
end


let () =
  ignore( Toolbar.combo_bin#connect#changed ~callback:Binary.callback ) ;
  ignore( Toolbar.combo_source#connect#changed ~callback:Source.callback )

let dummy_text = "\000\000\000\000"
let refresh () =
  Source.combo_callback_last_text := dummy_text;
  Source.callback_now := true;
  Source.callback () ;

  Binary.combo_callback_last_text := dummy_text;
  Binary.callback_now := true;
  Binary.callback ()
