open Pipe

let (//) = Filename.concat

module Const = struct
  let executable_root = Filename.dirname (Filename.dirname Util.executable_name)

  let cygwin_setup_url =
    match Sys.os_type with
    | "Win32" -> "http://cygwin.com/setup.exe"
    | _ -> "http://localhost/godi/setup.exe"

  let cygwin_mirror_list = "http://cygwin.com/mirrors.lst"

  let preferred_mirrors = [| "ftp.gwdg.de" ; "cygwin.cict.fr" |]

  (* this list will only be used, if the download/parsing of cygwin_mirror_list fails *)
  let cygwin_default_mirrors = [|
    "http://artfiles.org/cygwin.org/pub/cygwin/";
    "http://box-soft.com/";
    "http://cygwin.cict.fr/";
    "http://cygwin.cybermirror.org/";
    "http://cygwin.mirror.rafal.ca/";
    "http://cygwin.mirrorcatalogs.com/";
    "http://cygwin.mirrors.hoobly.com/";
    "http://cygwin.mirrors.pair.com/";
    "http://cygwin.parentingamerica.com/";
    "http://cygwin.petsads.us/";
    "http://cygwin.skazkaforyou.com/";
    "http://cygwin.uib.no/";
    "http://cygwin.xl-mirror.nl/";
    "http://download.nus.edu.sg/mirror/cygwin/";
    "http://ftp-stud.hs-esslingen.de/pub/Mirrors/sources.redhat.com/cygwin/";
    "http://ftp.cc.uoc.gr/mirrors/cygwin/";
    "http://ftp.eq.uc.pt/software/pc/prog/cygwin/";
    "http://ftp.fsn.hu/pub/cygwin/";
    "http://ftp.gwdg.de/pub/linux/sources.redhat.com/cygwin/";
    "http://ftp.iij.ad.jp/pub/cygwin/";
    "http://ftp.inf.tu-dresden.de/software/windows/cygwin32/";
    "http://ftp.jaist.ac.jp/pub/cygwin/";
    "http://ftp.rediris.es/mirror/cygwin/";
    "http://ftp.rnl.ist.utl.pt/pub/cygwin/";
    "http://ftp.yz.yamagata-u.ac.jp/pub/cygwin/";
    "http://linux.rz.ruhr-uni-bochum.de/download/cygwin/";
    "http://lug.mtu.edu/cygwin/";
    "http://mirror.aarnet.edu.au/pub/sourceware/cygwin/";
    "http://mirror.cpsc.ucalgary.ca/mirror/cygwin.com/";
    "http://mirror.cs.vt.edu/pub/cygwin/cygwin/";
    "http://mirror.csclub.uwaterloo.ca/cygwin/";
    "http://mirror.isoc.org.il/pub/cygwin/";
    "http://mirror.mcs.anl.gov/cygwin/";
    "http://mirror.steadfast.net/cygwin/";
    "http://mirror.switch.ch/ftp/mirror/cygwin/";
    "http://mirror.symnds.com/software/cygwin/";
    "http://mirror.team-cymru.org/cygwin/";
    "http://mirrors.163.com/cygwin/";
    "http://mirrors.fe.up.pt/pub/cygwin/";
    "http://mirrors.mojhosting.sk/cygwin/";
    "http://mirrors.syringanetworks.net/cygwin/";
    "http://mirrors.ustc.edu.cn/cygwin/";
    "http://mirrors.xmission.com/cygwin/";
    "http://repo.fedoramd.org/mirrors/cygwin/";
    "http://sourceware.mirrors.tds.net/pub/sourceware.org/cygwin/";
    "http://stingray.cyber.net.pk/pub/cygwin/";
    "http://ucmirror.canterbury.ac.nz/cygwin/";
    "http://www.mirrorservice.org/sites/sourceware.org/pub/cygwin/";
                               |]

  (* a special make build, that can handle rules with windows path:
     C:/xyz/test.o:
         ....
  *)
  let make2url = "http://www.cmake.org/files/cygwin/make.exe-cygwin1.7"
  let make2location = (Filename.dirname executable_root) // "sbin" // "make2.exe"

  let packages =
    (match Sys.word_size with
    | 32 -> "mingw64-i686-gcc-core"
    | 64 -> "mingw64-x86_64-gcc-core"
    | _ -> assert false)
    ^ ",make,patch,bc,cpio,rlwrap,zsh"
    ^ ",xz,gawk,less,bzip2,diffutils,dos2unix" (* this line is unneccessary at the moment *)

  (* default location gui/bin/installer.exe *)
  let installer_root_mixed = Util.windows_mixed_path ^$
    Filename.dirname ^$
    Filename.dirname ^$
    Filename.dirname Util.executable_name

  let pixbuf_caml =
    let pixbuf_name = "camel96.png" in
    try
      let img = executable_root // "share" // "pixmaps" // pixbuf_name in
      if Sys.file_exists img then
        GdkPixbuf.from_file img
      else
        let img = executable_root // "pixmaps" // pixbuf_name in
        if Sys.file_exists img then
          GdkPixbuf.from_file img
        else
          raise Not_found
    with
    | _ -> GdkPixbuf.create ~width:1 ~height:1 ()

  let help_dir ~var ~path =
    try Sys.is_directory ( (Sys.getenv var) // path )
    with
    | Not_found
    | Sys_error _ -> false

  let reg_exe_found =
    try
      ignore ( Glib.find_program_in_path "reg.exe" );
      true
    with
    | Not_found -> false

  let start_menu_dir_found =
    (* Windows XP, perhaps also works for newer versions *)
    ( reg_exe_found && ( try ignore ( Sys.getenv "USERPROFILE" ) ; true
      with Not_found -> false ) )
    (* Windows Vista or newer, default *)
    || ( help_dir ~var:"APPDATA" ~path:(
      "Microsoft" // "Windows" // "Start menu") )

  let desktop_dir_found = help_dir ~var:"USERPROFILE"
    ~path:"Desktop"

  let godi_dir_name = Printf.sprintf "wodi%d" Sys.word_size
end


type cgwin_install =
| Cyg_complete
| Cyg_update
| Cyg_skip

module Gl = struct
  let ocaml_to_path = (Const.reg_exe_found, ref false)
  let cygwin_to_path = (Const.reg_exe_found,ref false)
  let set_env_var = (Const.reg_exe_found,ref false)
  let sc_start_menu = (Const.start_menu_dir_found, ref false)
  let sc_desktop = (Const.desktop_dir_found, ref false)

  let cygwin_mirrors = ref Const.cygwin_default_mirrors
  let cygwin_mirror_to_use = ref Const.cygwin_default_mirrors.(0)
  let mirrors_fetched = ref false

  let cygwin_todo = ref Cyg_complete

  let folder = ref (match Sys.os_type with
  | "Win32" ->
    (match Sys.word_size with
    | 32 -> "C:\\wodi32"
    | _  -> "C:\\wodi64")
  | _ -> "/tmp/test" )
end

module Cygwin_mirrors = struct
  let get_host_from_http_url str =
    let last = String.index_from str 8 '/' in
    String.sub str 7 (last - 7)

  let http_mirror_regex = Str.regexp "^http://\\([^/]+\\)/\\([^;]*\\);[^;]+;[^;]+;[^;]+$"

  let extract_mirror line =
    if Str.string_match http_mirror_regex line 0 then
      Some("http://" ^ (Str.matched_group 1 line) ^ "/" ^ (Str.matched_group 2 line))
    else
      None

  let init () =
    let run () =
      let f accu line =
        match extract_mirror line with
        | None -> accu
        | Some(site) -> site::accu
      in
      try
        let mirrors =
          let p = (new Ecurl.ecurl ~use_cookies:false ())#get Const.cygwin_mirror_list in
          Util.split p '\n'
          |> List.fold_left f []
          |> List.rev
          |> Array.of_list
        in
        match mirrors with
        | [| |] ->
          GtkThread.async ( fun a -> a := true ) Gl.mirrors_fetched
        | _  ->
          let f a =
            Gl.cygwin_mirrors := a ;
            Gl.mirrors_fetched := true
          in
          GtkThread.async f mirrors;
      with
      | _ -> GtkThread.async ( fun a -> a := true ) Gl.mirrors_fetched
    in
    let _ = Thread.create run () in
    ()
end




(*let invalid_string = "\000"*)

let () =
  ignore ( Glib.Timeout.add ~ms:100 ~callback:GtkThread.do_jobs )

let show_error ~title ~message =
  Info_simple.show_error ~width:320 ~height:200 ~window_title:title  ~title ~message ()

let ask ~title ~message =
  Info_simple.ask ~width:320 ~height:200 ~window_title:title  ~title ~message ()


(* file should be empty and writable, try to create it, if it doesn't exist already *)
let check_dir x =
  if Sys.file_exists x = true then
    if Sys.is_directory x = false then
      `Error (Printf.sprintf "%s is not a directory" x)
    else
      try
        match Sys.readdir x with
        | [| |] ->
          let dummy = Filename.concat x "dummy.txt" in
          (try
             let ch = open_out dummy in
             Util.try_finally ( fun ch ->
               output_string ch "test"
             ) ch ( fun ch ->
               close_out ch;
               Sys.remove dummy ) ch ;
             `Ok
           with
           | _ -> `Error (Printf.sprintf "can't create files in %s" x)  )
        | _ -> `Error (Printf.sprintf "%s already contains files" x)
      with
      | Sys_error e -> `Error (Printf.sprintf "I can't access %s (%s)" x e)
  else
    match try Unix.mkdir x 0o755 ; true with _ -> false with
    | false -> `Error (Printf.sprintf "creation of '%s' failed" x)
    | true  -> `Ok


let check_cygwin x =
  let dir = x // "bin" in
  if Sys.file_exists dir = false ||  Sys.is_directory dir = false then
    `Nocygwin
  else if Sys.file_exists (dir // "cygpath.exe") = false || Sys.file_exists (dir // "passwd.exe" ) = false then
    `Nocygwin
  else if Sys.file_exists (x // "opt" // Const.godi_dir_name ) then
    `Wodi_already_installed
  else
    let files =
      let mp = match Sys.word_size with
      | 32 -> "i686-w64-mingw32-"
      | 64 -> "x86_64-w64-mingw32-"
      | _ -> assert false
      in
      [| "ash.exe" ; "basename.exe" ; "bash.exe" ; "bc.exe" ; "bunzip2.exe" ; "cat.exe" ; "chmod.exe" ;
         "chown.exe" ; "cmp.exe" ; "cp.exe" ; "cpio.exe" ; "cut.exe" ; "cygcheck.exe" ; "dash.exe" ;
         "date.exe" ; "dd.exe" ; "diff.exe" ; "dirname.exe" ; "dos2unix.exe" ; "echo.exe" ; "env.exe" ;
         "file.exe" ; "find.exe" ; "gawk.exe" ; "grep.exe" ; "gzip.exe" ; "head.exe" ; "install.exe" ;
         "less.exe" ; "ln.exe" ; "ls.exe" ; "make.exe" ; "man.exe" ; "mintty.exe" ; "mkdir.exe" ;
         "mktemp.exe" ; "mount.exe" ; "mv.exe" ; "patch.exe" ; "pwd.exe" ; "readlink.exe" ; "rlwrap.exe" ;
         "rm.exe" ; "rmdir.exe" ; "sed.exe" ; "sh.exe" ; "sort.exe" ; "split.exe" ; "stat.exe" ; "tail.exe" ;
         "tar.exe" ; "test.exe" ; "touch.exe" ; "tr.exe" ; "xargs.exe" ; "xz.exe" ; "yes.exe" ;
         mp ^ "ar.exe"; mp ^ "as.exe" ; mp ^ "gcc.exe" ; mp ^ "ld.exe" ; mp ^ "nm.exe" ; mp ^ "ranlib.exe" ;
         mp ^ "strings.exe" ; mp ^ "strip.exe"
      |]
    in
    let f file = Sys.file_exists (dir // file) = false  in
    if Util.array_exists f files then
      `Incomplete_cygwin
    else
      `Complete_cygwin


let install_all cygwin_prog =
  let bash = !Gl.folder // "bin" // "bash.exe"
  and godi_dir_cygwin = "/opt/" ^ Const.godi_dir_name
  and godi_dir_win = !Gl.folder // "opt" // Const.godi_dir_name in
  let godi_console =  godi_dir_cygwin ^ "/sbin/godi_console.exe"
  and godi_env =
    let cur_path =
      try Sys.getenv "PATH" with Not_found ->
        try Sys.getenv "Path" with Not_found -> ""
    in
    Array.append
      (Util.environment_with_removed_values ["PATH";"Path"])
      [| "CYGWIN=nodosfilewarning";
         "PATH=" ^
           ( godi_dir_win // "bin" ) ^ ";" ^
           ( godi_dir_win // "sbin" ) ^ ";" ^
           ( !Gl.folder // "bin") ^ ";" ^
           cur_path  |]
  (* install cygwin *)
  and c_cygwin = {
    Console.command =  cygwin_prog ;
    Console.args = [| "--quiet-mode";
                      "--no-shortcuts";
                      "--root";
                      !Gl.folder;
                      "--local-package-dir";
                      Filename.concat (Glib.get_tmp_dir ()) "cygwin";
                      "--site";
                      !Gl.cygwin_mirror_to_use;
                      "--categories=Base";
                      "--packages=" ^ Const.packages ;
                   |]  ;
    Console.killable = true;
    Console.description = "Downloading and installing various cygwin packages";
    Console.env = None;
  }
  in
  let c_help args desc =
    let rec quote_needed str i =
      if i < 0 then false
      else
        match str.[i] with
        | '\\'
        | '\"'
        | ' '
        | '\t'
        | '\n'
        | '\x0B' -> true (*vercial tab*)
        | _ -> quote_needed str (pred i)
    in
    let f str =
      let len = String.length str in
      match quote_needed str (len - 1 ) with
      | false -> str
      | true  ->
        let buf = Buffer.create (len + 5) in
        Buffer.add_char buf '\"';
        for i = 0 to len - 1 do
          let cur = str.[i] in
          (match cur with
          | '\"' | '\\' -> Buffer.add_char buf '\\';
          | _ -> ());
          Buffer.add_char buf cur;
        done;
        Buffer.add_char buf '\"';
        Buffer.contents buf;
    in
    let args = List.map f args in
    let command = String.concat " " args in
    {
      Console.command =  bash ;
      Console.args = [| "--login"; "-c" ; command |];
      Console.killable = true;
      Console.description = desc;
      Console.env = Some godi_env;
    }
  in

  let c_mkdir = c_help
    [ "mkdir" ; "-p" ; godi_dir_cygwin ]
    "Creating godi folder"
  and c_copy = c_help
    [ "cp" ; "-a" ; Const.installer_root_mixed ^ "/." ; godi_dir_cygwin ]
    "Copying files"
  and c_copy2 = c_help
    [ "cp" ; cygwin_prog ; "/setup.exe" ]
    "Copying cygwin's setup.exe"
  and c_relocate = c_help
    [ godi_dir_cygwin ^ "/gui/bin/relocate.sh" ; godi_dir_cygwin ]
    "Changing godi configuration"
  and c_add_ocaml = c_help
    [ godi_console ; "add" ; "-v" ; "godi-ocaml" ]
    "Installing godi-ocaml"
  and c_add_flexdll = c_help
    [ godi_console ; "add" ; "-v" ; "godi-flexdll" ]
    "Installing godi-flexdll"
  and c_add_findlib = c_help
    [ godi_console ; "add" ; "-v" ; "godi-findlib" ]
    "Installing godi-findlib"
  and c_winconfig =
    let help (_,{contents}) switch l = match contents with
    | true -> switch :: l
    | false -> l
    in
    let args = ["--first-run" ; godi_dir_cygwin ^ "/lib/godi/winconfig.sh"]
      |> help Gl.ocaml_to_path "--add-ocaml-path"
      |> help Gl.cygwin_to_path "--add-cygwin-path"
      |> help Gl.set_env_var "--global-environment"
      |> help Gl.sc_start_menu "--add-startmenu-folder"
      |> help Gl.sc_desktop "--add-to-desktop"
      |> List.rev
    in
    c_help args
      "Configuring your installation"
  in
  let all = [ c_cygwin ; c_mkdir ; c_copy ; c_copy2 ; c_relocate ; c_add_ocaml ;
              c_add_flexdll ; c_add_findlib ; c_winconfig ]
  in
  match Console.run ~icon:Const.pixbuf_caml ~auto_close:true ~window_title:"Installing Wodi ... " all with
  | Console.Error _
  | Console.Aborted _ -> GMain.quit ()
  | Console.Ok ->
    let message =
      let buf = Buffer.create 256 in
      Buffer.add_string buf "The installation is now finished.\n" ;
      let desktop = !(snd Gl.sc_desktop)
      and startm = !(snd Gl.sc_start_menu) in

      if desktop = false && startm = false then (
        Buffer.add_string buf "You can start the Wodi manager with '";
        Buffer.add_string buf !Gl.folder;
        Buffer.add_string buf "\\Wodi";
        Buffer.add_string buf (string_of_int Sys.word_size);
        Buffer.add_string buf " Package\ Manager.lnk'. Or open a cygwin shell with '";
        Buffer.add_string buf !Gl.folder;
        Buffer.add_string buf "\\Wodi";
        Buffer.add_string buf (string_of_int Sys.word_size);
        Buffer.add_string buf " Cygwin.lnk'."
      )
      else (
        Buffer.add_string buf "You can start the Wodi Manager or open a cygwin \
shell by following the links ";
        if desktop = true && startm = true then
          Buffer.add_string buf "in your start menu or on your desktop."
        else if desktop = true then
          Buffer.add_string buf "on your desktop."
        else
          Buffer.add_string buf "in your start menu."
      );
      Buffer.contents buf
    and title = "Installation finished!" in
    Info_simple.show_info ~width:420 ~height:230 ~window_title:title  ~title ~message () ;
    GMain.quit ()


let rec ask_for_file parent =
  let dialog = GWindow.file_chooser_dialog
    ~action:`SELECT_FOLDER
    ~title:"Open File"
    ~parent ()
  in
  dialog#add_button_stock `CANCEL `CANCEL ;
  dialog#add_select_button_stock `OPEN `OPEN ;
  let erg = match dialog#run () with
  | `OPEN -> dialog#filename
  | `DELETE_EVENT | `CANCEL -> None
  in
  dialog#destroy ();
  match erg with
  | None -> None
  | Some x -> check_directory parent x
and check_directory w x =
  Gl.cygwin_todo := Cyg_complete;
  let message = match Util.contains_spaces x with
  | true -> `Error "Your path contains spaces.\nPlease choose another one!"
  | false ->
    match check_cygwin x with
    | `Nocygwin -> check_dir x
    | `Wodi_already_installed -> `Error "Wodi already installed. Please choose another path!"
    | `Incomplete_cygwin -> Gl.cygwin_todo := Cyg_update ; `Ask
    | `Complete_cygwin -> Gl.cygwin_todo := Cyg_skip ; `Ok
  in
  match message with
  | `Ok -> Some x
  | `Error message ->
    show_error ~title:"invalid path" ~message ;
    Gl.cygwin_todo := Cyg_complete;
    ask_for_file w
  | `Ask ->
    let title= "Incomplete cygwin installation"
    and message = "Your cygwin installation is incomplete. Abort, if you want to choose another path; otherwise your cygwin installation will be updated." in
    match ask ~title ~message with
    | true -> Some x
    | false -> Gl.cygwin_todo := Cyg_complete; ask_for_file w


let page0 w (vbox:GPack.box) set_senstive =
  let _ = GMisc.label
    ~markup:"<big><b>Thank You for Choosing WODI!</b></big>\n\n\
Please select an installation path. The path should not contain any \
whitespace characters."
    ~ypad:7
    ~xalign:0.
    ~line_wrap:true
    ~packing:vbox#pack
    ()
  in
  let _ =
    let hbox =
      GPack.hbox ~homogeneous:false ~packing:vbox#pack ~spacing:5 ()
    in
    let _ = GMisc.label ()
      ~text:"Path: "
      ~packing:hbox#pack
    in
    let e = GEdit.entry ~editable:false ~text:!Gl.folder ()
      ~packing:hbox#pack
    in
    let b = GButton.button (*~stock:`OPEN*) ~label:"change" ()
      ~packing:hbox#pack
    in
    b#set_image (GMisc.image ~stock:`OPEN ())#coerce ;
    let change_dir () =
      match ask_for_file w with
      | Some x when x <> !Gl.folder ->
        Gl.folder := x;
        set_senstive ();
        e#set_text x
      | None
      | Some _ ->  ()
    in
    let _ = b#connect#clicked ~callback:change_dir in
    e#misc#set_can_focus false
  in
  let _ =
    GMisc.label ~xalign:0. ~markup:"This location will contain a customized \
cygwin environment. Wodi will be installed in this environment.\n
If you want to use an existing Cygwin environment, select its root folder. \
If something is missing from your Cygwin installation, it will be upgrated.\n\n
<a href=\"http://ocaml.org/\">OCaml</a>, \
<a href=\"http://projects.camlcity.org/projects/findlib.html\">Findlib</a>, and \
<a href=\"http://alain.frisch.fr/flexdll.html\">FlexDLL</a> will be installed by default. \
Additional packages can be installed through the package management tool."
      ~line_wrap:true
      ~ypad:7
      ~packing:vbox#pack ()
  in
  ()


let page1 (vbox:GPack.box) =
  let _ = GMisc.label
    ~markup:"<big><b>Environment settings</b></big>\n\n\
Enable the checkboxes, if you want to modify the environment variables \
also outside cygwin. This is neccessary for some IDEs like \
<a href=\"http://www.algo-prog.info/ocaide/\">OcaIDE</a>."
    ~ypad:7
    ~xalign:0.
    ~line_wrap:true
    ~packing:vbox#pack
    ()
  in
  let () =
    let table = GPack.table ~rows:5 ~columns:2  ~packing:vbox#pack ()
    and tc = ref 0 in
    let add text (sensitive_state,var) =
      let i = !tc
      and cb = GButton.check_button ~active:!var ()
      and l = GMisc.label ~xalign:0. ~line_wrap:true ~text ()
      and callback () = var := not !var in
      cb#misc#set_sensitive sensitive_state;
      l#misc#set_sensitive sensitive_state;
      table#attach ~top:i ~left:1 cb#coerce;
      table#attach ~top:i ~left:2 l#coerce;
      ignore ( cb#connect#toggled ~callback );
      incr tc;
    in
    add "Add ocaml folder to %PATH%" Gl.ocaml_to_path ;
    add "Add cygwin folder to %PATH%" Gl.cygwin_to_path;
    add "Set environment variables (OCAMLLIB,..)" Gl.set_env_var;
    add "Create shortcuts in the start menu" Gl.sc_start_menu;
    add "Create desktop shortcuts" Gl.sc_desktop
  in
  ()


let get_default_mirror hosts =
  let i = ref 0
  and j = ref 0
  and n = ref (-1) in
  while !n < 0 && !i < Array.length Const.preferred_mirrors do
    while !n < 0 && !j < Array.length hosts do
      if Const.preferred_mirrors.(!i) = hosts.(!j) then
        n := !j ;
      incr j
    done;
    incr i
  done;
  if !n < 0 then
    (Array.length hosts) / 2
  else
    !n

let page2 (bnext:GButton.button) (vbox:GPack.box) =
  let dl_info = GMisc.label
    ~markup:"Downloading mirror list"
    ~xalign:0.
    ~line_wrap:true
    ()
  in
  dl_info#misc#set_sensitive false;
  let next im =
    if im then (
      bnext#misc#set_sensitive true;
      dl_info#misc#hide ();
    );
    let hosts =
      Array.map Cygwin_mirrors.get_host_from_http_url !Gl.cygwin_mirrors
    in
    let active = get_default_mirror hosts
    and strings = Array.to_list hosts
    and use_markup = false in
    Gl.cygwin_mirror_to_use := !Gl.cygwin_mirrors.(active);
    let (sl,_) = GEdit.combo_box_text ~use_markup ~active ~strings () in
    vbox#pack sl#coerce ;
    let callback () =
      let i = sl#active in
      Gl.cygwin_mirror_to_use := !Gl.cygwin_mirrors.(i);
    in
    let _ = sl#connect#changed ~callback in
    ()
  in
  let callback () =
    match !Gl.mirrors_fetched  with
    | false -> true
    | true ->
      next false;
      false
  in
  let _ = GMisc.label
    ~markup:"<big><b>Cygwin Download</b></big>\n\n\
The configuration is now finished. \
During the next step, cygwin will be downloaded and installed. \
If the installation of cygwin fails, please try again, but choose \
another mirror from below."
    ~ypad:7
    ~xalign:0.
    ~line_wrap:true
    ~packing:vbox#pack
    ()
  in
  match !Gl.mirrors_fetched with
  | false ->
    bnext#misc#set_sensitive false;
    vbox#pack dl_info#coerce;
    let _ = Glib.Timeout.add ~ms:500 ~callback in
    ()
  | true -> next true


(*
let http_mirror_regex = Str.regexp "http://\\([^/]+\\)/\\([^;]*\\);\\([^;]+\\);"
let extract_mirror line =
  if Str.string_match http_mirror_regex line 0 then
    Some("http://" ^ (Str.matched_group 1 line) ^ "/" ^ (Str.matched_group 2 line), (Str.matched_group 3 line))
  else
    None
;;
extract_mirror "http://ftp.fsn.hu/pub/cygwin/;ftp.fsn.hu;Europe;Hungary" ;;

*)
(* http://ftp.fsn.hu/pub/cygwin/;ftp.fsn.hu;Europe;Hungary *)





let download_cygwin () =
  let error = ref false in
  let tmpfile =
    try
      Filename.temp_file ~temp_dir:(Glib.get_tmp_dir ())
            "setup" ".exe"
    with
    | x ->
      let x = match x with
      | Sys_error x -> x
      | _ -> "unknown error"
      in
      error:= true;
      Printf.sprintf "I can't create a temporary file (%s). \
This error is fatal, the program will terminate now." x
  in
  match !error with
  | true ->
    show_error ~title:"Fatal error" ~message:tmpfile ;
    GMain.quit ()
  | false ->
    match Download_dialog.dialog [(Const.cygwin_setup_url,tmpfile);(Const.make2url,Const.make2location)]
    ~window_title:"Downloading cygwin's setup.exe and cmake's make.exe"
    with
    | false ->
      let message = "The download of cygwin's 'setup.exe' failed. \
This error is fatal, I can't continue." in
      show_error ~title:"Fatal error" ~message;
      GMain.quit ()
    | true ->
      install_all tmpfile


let main () =
  let next_action_dl = ref false in
  let quit_or_next () =
    match !next_action_dl with
    | true -> download_cygwin ()
    | false -> GMain.quit ()
  in


  let width = 450
  and height = 300 in

  let w = GWindow.dialog ~border_width:6  ~title:"Wodi Installation" ~position:`CENTER ()  in
  w#set_default_size ~width ~height;
  ignore ( w#connect#destroy ~callback:quit_or_next );


  let image = GMisc.image ~xalign:0. ~yalign:0. ~xpad:15 ~ypad:25 ~pixbuf:Const.pixbuf_caml () in
  w#set_icon (Some Const.pixbuf_caml);

  let hbox = GPack.hbox () in
  w#vbox#pack ~expand:true ~fill:true hbox#coerce;

  hbox#pack image#coerce;

  let vbox = GPack.vbox () in
  hbox#pack vbox#coerce;

  let vbox_content = GPack.vbox () in
  vbox#pack ~expand:true ~fill:true vbox_content#coerce;

  let next_button =
    let hbox2 = GPack.hbox ~packing:vbox#pack () in
    (* dummy label to expand *)
    let l = GMisc.label ~selectable:false ~text:""
      ~packing:(hbox2#pack ~expand:true ~fill:true) ()
    in
    l#misc#set_can_focus false ;

    (* real hbox *)
    let hbox3 = GPack.hbox ~homogeneous:true
      ~packing:(hbox2#pack) ()
    in
    let c_b = GButton.button ~stock:`CANCEL ~packing:hbox3#pack () in

    let ok_b = GButton.button ~use_mnemonic:true  ~label:"N_ext" ()
      ~packing:hbox3#pack
    in
    ok_b#set_image (GMisc.image ~stock:`GO_FORWARD ())#coerce ;
    ok_b#misc#set_sensitive false;
    ignore ( c_b#connect#clicked ~callback:quit_or_next );
    ok_b
  in

  let set_sensitive () = next_button#misc#set_sensitive true
  and cur_page = ref 0 in
  let rec next () =
    match !cur_page with
    | 0 ->
      incr cur_page;
      set_sensitive ();
      page0 w vbox_content set_sensitive;
    | 1 ->
      begin match check_directory w !Gl.folder with
      | None -> ()
      | Some x  ->
        Gl.folder := x;
        incr cur_page;
        List.iter vbox_content#remove vbox_content#children ;
        page1 vbox_content
      end;
    | 2 ->
      incr cur_page;
      if !Gl.cygwin_todo = Cyg_skip then
        next ()
      else (
        next_button#set_label "I_nstall";
        next_button#set_image (GMisc.image ~stock:`APPLY ())#coerce ;
        List.iter vbox_content#remove vbox_content#children ;
        page2 next_button vbox_content
      )
    | _ ->
      next_action_dl:=true;
      w#destroy ();
  in
  ignore ( next_button#connect#clicked ~callback:next );
  next ();
  w#show ();
  Cygwin_mirrors.init ();
  GtkThread.main ()


let () =
  main ()
