type run = {
  command: string;
  args: string array;
  killable: bool ; (* whether we should kill the process or not if something goes wrong *)
  description: string;
  env: string array option;
}

type t =
| Ok
| Error of run list
| Aborted of run list

(* ?chdir ~command ~args () *)
let run ?(width=540) ?(height=360) ?icon ?success_message ?(auto_close=false) ~window_title = function
| [] -> Ok
| run_list ->
  let w = GWindow.dialog  ~title:window_title ~modal:true ~position:`CENTER ()  in
  (match icon with
  | None -> ()
  | (Some _) as sx ->  w#set_icon sx );
  let s = GBin.scrolled_window ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:w#vbox#add  () in
  let t = GText.view ~editable:false ~wrap_mode:`CHAR ~border_width:4 ~width ~height
    ~cursor_visible:false ~packing:s#add () in
  t#misc#modify_bg [(`NORMAL,`BLACK)];
  t#misc#modify_base [(`NORMAL,`BLACK)];
  t#misc#modify_text [(`NORMAL,`WHITE)];
  t#misc#set_can_focus false ;
  let buff = GText.buffer () in
  let iter = buff#start_iter in

  let hbox = GPack.hbox ~border_width:0 ~spacing:0 ~homogeneous:true () in
  w#vbox#pack (GMisc.label ~yalign:0.5 ~xalign:0. ~selectable:false ~text:"" ())#coerce ;
  w#vbox#pack (*~expand:true ~fill:true*) hbox#coerce;

  let label = GMisc.label ~yalign:0.5 ~xalign:0.
    ~selectable:false ~show:true ~text:"" ~packing:hbox#pack
    ~line_wrap:true ()
  in
  let ali = GBin.alignment ~xalign:1. ~yalign:1. () in
  hbox#pack (*~expand:true ~fill:false*) ali#coerce;
  let hbox2 = GPack.hbox ~spacing:5 ~homogeneous:true ~packing:ali#add () in
  let ok_b = GButton.button ~stock:`OK ~packing:hbox2#add () in
  let c_b = GButton.button ~stock:`CANCEL ~packing:hbox2#add () in

  (*ok_b#misc#set_size_request ~width:100 ();
  c_b#misc#set_size_request ~width:100 (); *)

  ok_b#misc#set_sensitive false;

  t#set_buffer buff;
  ignore ( t#scroll_to_iter iter );

  let status = ref (Aborted [])
  and run_list = ref run_list
  and current = ref None
  and js = ref None
  and finished = ref false in

  ignore ( c_b#connect#clicked ~callback:( fun () -> finished := true;  w#response `DELETE_EVENT ) );
  ignore ( ok_b#connect#clicked ~callback:( fun () -> finished := true; w#response `DELETE_EVENT ) );


  let rec update_buffer str =
    if !finished = false then (
      buff#insert ~iter str;
      buff#place_cursor ~where:iter ;
      ignore ( t#scroll_to_mark ~use_align:true ~within_margin:0.0 ~yalign:1. ~xalign:1. `INSERT );
    )
  and a_err_b() =
    if !finished = false then (
      status := Error [];
      let text =match !current with
      | None -> " <span color=\"red\"><b> Something went wrong, but I don't know what </b> :-(</span>"
      | Some x -> " <span color=\"red\"><b>" ^ x.description ^ " failed </b> :-(</span>"
      in
      w#set_title window_title;
      label#set_text text;
      label#set_use_markup true;
    )
  and next () = match !run_list with
  | hd::tl ->
    if !finished = false then (
      run_list := tl;
      current := Some hd;
      let text = " " ^ hd.description ^ " ..." in
      label#set_text text;
      w#set_title text;
      let command = hd.command
      and args = hd.args
      and env = hd.env
      in
      let js' = Job.prepare ~f_output:update_buffer  ~f_finish_ok:next ~f_finish_err:a_err_b
        ?env ~command ~args () in
      js :=  (match hd.killable with
      | true -> Some js';
      | false -> None );
      Job.run js'
    )
  | [] ->
    if !finished = false then (
      status := Ok ;
      current := None;
      match auto_close with
      | true -> w#response `DELETE_EVENT
      | false ->
        w#set_title window_title;
        let text = match success_message with
        | None -> " <span color=\"green\"><b> Request fullfilled </b> :-)</span>"
        | Some x -> x
        in
        label#set_text text;
        label#set_use_markup true;
        ok_b#misc#set_sensitive true
    )
  in
  next () ;
  ignore (match w#run () with
  | `DESTROY -> ()
  | ` CANCEL | `DELETE_EVENT -> w#destroy ()
  | _ -> w#destroy () );
  (match !js with
  | Some js ->
    Job.stop js;
  | None -> () (* not killable *) );
  finished := true;
  match !status with
  | Ok -> Ok
  | Error _
  | Aborted _ ->
    let to_run = !run_list in
    let to_run = match !current with
    | None -> to_run
    | Some x  -> x::to_run
    in
    match !status with
    | Error _ -> Error to_run
    | Aborted _ -> Aborted to_run
    | Ok -> assert false
