let remove_no_exn file = try Sys.remove file with | _ -> ()


let run_mutexed m f x =
  Mutex.lock m;
  Util.try_finally f x Mutex.unlock m

let create_mutex_flag start =
  let mutex = Mutex.create ()
  and flag = ref start in
  let f1 () = run_mutexed mutex ( fun x -> !x ) flag
  and f2 () = run_mutexed mutex ( fun x -> x:= not start) flag in
  f1,f2

let dialog ?(width=480) ?(height=140) ~window_title = function
| [] -> true
| dl_list ->
  let w = GWindow.dialog ~title:window_title ~modal:true ~position:`CENTER ()
  and recip_dl_list_length = 1. /. ( float_of_int (List.length dl_list)) in


  let pixbuf =  w#misc#render_icon ~size:`DIALOG `DIALOG_INFO in
  let image = GMisc.image ~xpad:10 ~ypad:10 ~xalign:0. ~yalign:0. ~pixbuf () in
  w#set_icon (Some pixbuf);


  let hbox = GPack.hbox ~border_width:5 ~spacing:5 () in
  w#vbox#pack hbox#coerce;

  hbox#add image#coerce;

  let vbox = GPack.vbox ~border_width:5 ~spacing:5 () in
  hbox#pack ~expand:true ~fill:true vbox#coerce;

  vbox#misc#set_size_request ~height:(height - 20) ~width:(width - 40) () ;

  (*let text = "Downloading " in*)

  let desc_label = GMisc.label ~text:"Downloading " ~xalign:0. ~yalign:0.5  ~packing:vbox#pack () in

  let table = GPack.table ~rows:2 ~columns:2 ~packing:vbox#pack () in


  let _ = GMisc.label ~text:"Current: " ~xalign:0. ~yalign:0.5 ()
    ~packing:(table#attach ~left:0 ~right:1 ~top:0 )
  in

  let pbar_current = GRange.progress_bar ~show:false ()
    ~packing:(table#attach ~left:1 ~right:2 ~top:0 ~expand:`BOTH ~fill:`X ~shrink:`BOTH)
  in
  pbar_current#set_fraction 0.000001 ;
  pbar_current#misc#show ();

  let _ = GMisc.label ~text:"Whole: " ~xalign:0. ~yalign:0.5 ()
    ~packing:(table#attach ~left:0 ~right:1 ~top:1 )
  in

  let pbar_whole = GRange.progress_bar ~show:false ()
    ~packing:(table#attach ~left:1 ~right:2 ~top:1 ~expand:`BOTH ~fill:`X ~shrink:`BOTH)
  in
  pbar_whole#set_fraction 0.000001 ;
  pbar_whole#misc#show ();

  let is_abort_set, set_abort = create_mutex_flag false in
  let regular_finished,set_regular_finished = create_mutex_flag false in

  (* I don't know how to make gtk happy *)
  let fra_whole = ref 0.00001 in

  let progress dltotal dlnow _ultotal _ulnow =
    match is_abort_set () with
    | true -> true
    | false ->
      let fra = dlnow /. dltotal in
      GtkThread.async ( fun fra ->
        pbar_current#set_fraction fra ;
        let cur = min ( !fra_whole +. ( fra *. recip_dl_list_length )  ) 0.999999 in
        pbar_whole#set_fraction cur;
      ) fra ;
      false
  in
  let tmp_file  = ref None in

  let start_time = Unix.gettimeofday () in

  let kill () =
    w#response `DELETE_EVENT ;
    false
  in
  let rec download = function
  | [] ->
    set_regular_finished ();
    w#response `DELETE_EVENT
  | (url,dat)::tl ->
     let iter =
     try
       let c = new Ecurl.ecurl () in
       let file = Filename.temp_file ~temp_dir:(Filename.dirname dat) (Filename.basename dat) ".tmp" in
       GtkThread.async ( fun url ->
         tmp_file := Some file;
         let file = match Ecurl.get_filename_from_url url with
         | None -> url
         | Some x -> x
         in
         desc_label#set_text ("Downloading " ^ file)  ) url;
       (try
          c#get2file ~fprogress:progress  ~file url ;
        with
        | x -> remove_no_exn file; raise x);
       remove_no_exn dat;
       Sys.rename file dat;
       GtkThread.async ( fun () ->
         fra_whole := !fra_whole +. recip_dl_list_length ;
         pbar_whole#set_fraction (min !fra_whole 0.999999)
       ) () ;
       true
    with
    | _ -> false
    in
    match iter with
    | true -> download tl
    | false ->
      let end_time = Unix.gettimeofday () in
      if end_time -. start_time < 3. then
        (* make sure this window doesn't vanish instantly.
           This would confuse the user *)
        ignore( Glib.Timeout.add ~ms:2000 ~callback:kill )
      else
        GtkThread.async w#response `DELETE_EVENT
  in
  w#add_button_stock `CANCEL `CANCEL;

  Util.try_finally ( fun () ->
    let _ = Thread.create download dl_list in
    (match w#run () with
    | _ -> w#destroy ())
  ) () set_abort () ;
  (match !tmp_file with
  | None -> ()
  | Some x -> remove_no_exn x);
  regular_finished ()
