#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/unixsupport.h>

#if _WIN32
static char *
get_executable_name(void)
{
#define START_LENGTH  1024
#define MAX_LENGTH    (START_LENGTH * 64)
    DWORD result;
    DWORD path_size = START_LENGTH;
    char* path      = malloc(START_LENGTH);

    while ( path != NULL && path_size <= MAX_LENGTH )
    {
        memset(path, 0, path_size);
        result = GetModuleFileName(NULL, path, path_size - 1);
        if (result == 0 )
        {
            free(path);
            path = NULL;
        }
        else if (result == path_size - 1)
        {
            DWORD er = GetLastError();
            free(path);
            path=NULL;
            if ( er == ERROR_INSUFFICIENT_BUFFER ||
                 er == ERROR_SUCCESS )
            {
                path_size = path_size * 2;
                if ( path_size <= MAX_LENGTH )
                  path = malloc(path_size);
            }
        }
        else
        {
            /* don't waste memory */
            size_t len = 1 + strlen(path);
            char * result = malloc(len);
            if ( result )
              result=memcpy(result, path, len);
            free(path);
            path=result;
            path_size= MAX_LENGTH + 1;
        }
    }
    return path;
#undef START_LENGTH
#undef MAX_LENGTH
}
#endif


CAMLprim value exec_name(value unit);

CAMLprim value exec_name(value unit)
{
    CAMLparam0();
    CAMLlocal1(ret);
#ifdef _WIN32
    char * str = get_executable_name();
    if ( str == NULL ){
        win32_maperr(GetLastError());
        uerror("exec_name", Nothing);
    }
    ret=caml_copy_string(str);
    free(str);
#else
    caml_failwith("exec_name is only avaiable on windows");
#endif
    CAMLreturn(ret);
}
