type t =
| Info
| Error
| Question

let run ?(width=480) ?(height=320) ~typ' ~window_title ~title ~message  () =
  let w = GWindow.dialog ~title:window_title ~modal:true ~position:`CENTER ()  in

  let pixbuf = match typ' with
  | Error -> w#misc#render_icon ~size:`DIALOG `DIALOG_ERROR
  | Question ->  w#misc#render_icon ~size:`DIALOG `DIALOG_QUESTION
  | Info ->  w#misc#render_icon ~size:`DIALOG `DIALOG_INFO
  in
  let image = GMisc.image ~xpad:10 ~ypad:10 ~xalign:0. ~yalign:0. ~pixbuf () in
  w#set_icon (Some pixbuf);

  let hbox = GPack.hbox ~border_width:5 ~spacing:5 () in
  w#vbox#pack hbox#coerce;

  hbox#add image#coerce;

  let vbox = GPack.vbox ~border_width:5 ~spacing:5 () in
  hbox#pack ~expand:true ~fill:true vbox#coerce;

  let title = "<b><big>" ^ title ^ "</big></b>" in
  let label_title = GMisc.label ~yalign:0. ~xalign:0. ~selectable:false ~markup:title
    ~line_wrap:true ~packing:vbox#pack () in
  label_title#misc#set_can_focus false ;

  let frame = GBin.frame ~border_width:3 ~shadow_type:`IN ~packing:vbox#add () in
  let s = GBin.scrolled_window ~hpolicy:`NEVER ~vpolicy:`AUTOMATIC ~packing:frame#add  () in
  let swidth = width - 40 in
  let () =
    let vp = GBin.viewport ~shadow_type:`NONE ~border_width:0 ~packing:s#add ~show:true () in
    let label = GMisc.label ~line_wrap:true ~yalign:0. ~xalign:0. ~selectable:false ~text:message ~packing:vp#add () in
    label#misc#set_can_focus false ;
    label#misc#set_size_request ~width:(swidth - 5) () ;
  in
  s#misc#set_size_request ~height:(height - 20) ~width:swidth  () ;

  (*let frame = GBin.frame ~border_width:3 ~shadow_type:`IN ~packing:vbox#add () in
  let s = GBin.scrolled_window ~hpolicy:`NEVER ~vpolicy:`AUTOMATIC ~packing:frame#add  () in
  let () =
    let vp = GBin.viewport ~shadow_type:`NONE ~border_width:0 ~packing:s#add ~show:true () in
    let label = GMisc.label ~line_wrap:true ~yalign:0. ~xalign:0. ~selectable:false ~text:message ~packing:vp#add () in
    label#misc#set_can_focus false ;
  in
  s#misc#set_size_request ~height:(height - 20) ~width:(width - 40) () ; *)

  (match typ' with
  | Question -> w#add_button_stock `CANCEL `CANCEL
  | Info | Error -> () );

  w#add_button_stock `OK `OK;

  let ret = ref false in
  begin match w#run () with
  | `OK -> ret:=true ; w#destroy ()
  | _ -> w#destroy ()
  end;
  !ret


let show_error ?width ?height ~window_title ~title ~message () =
  let _ = run ~typ':Error ?width ?height ~window_title ~title ~message () in
  ()

let show_info ?width ?height ~window_title ~title ~message () =
  let _ = run ~typ':Info ?width ?height ~window_title ~title ~message () in
  ()

let ask ?width ?height ~window_title ~title ~message () =
  run ~typ':Question ?width ?height ~window_title ~title ~message ()
