let cols = new GTree.column_list
let col_text = cols#add Gobject.Data.string
let col_b = cols#add Gobject.Data.boolean

let fill_store (store:GTree.tree_store) tree =
  let f = function
  | (_,_,[]) -> ()
  | (s,b,l) ->
    let row = store#append () in
    store#set ~row ~column:col_text s ;
    store#set ~row ~column:col_b b ;
    let f s =
      let row = store#append ~parent:row () in
      store#set ~row ~column:col_text s;
      store#set ~row ~column:col_b false;
    in
    List.iter f l
  in
  List.iter f tree

let view_info =
  let renderer = GTree.cell_renderer_text [] in
  let cell_data_func (model:GTree.model) (row:Gtk.tree_iter) =
    let text = model#get ~row ~column:col_text  in
    let bold = match model#get ~row ~column:col_b with
    | true -> [`WEIGHT `ULTRABOLD]
    | false -> [`WEIGHT `NORMAL]
    in
    renderer#set_properties ( (`TEXT text ) :: bold )
  in
  let col = GTree.view_column () in
  col#set_sizing `FIXED;
  col#pack renderer ;
  col#set_cell_data_func renderer cell_data_func;
  col

let run ?(width=540) ?(height=360) ~is_error ~tree ~window_title ~title ~info () =
  let w = GWindow.dialog ~title:window_title ~modal:true ~position:`CENTER ()  in

  let pixbuf = match is_error with
  | true -> w#misc#render_icon ~size:`DIALOG `DIALOG_ERROR
  | false ->  w#misc#render_icon ~size:`DIALOG `DIALOG_QUESTION
  in
  let image = GMisc.image ~xpad:10 ~ypad:10 ~xalign:0. ~yalign:0. ~pixbuf () in
  w#set_icon (Some pixbuf);

  let hbox = GPack.hbox ~border_width:5 ~spacing:5 () in
  w#vbox#pack hbox#coerce;

  hbox#add image#coerce;

  let vbox = GPack.vbox ~border_width:5 ~spacing:5 () in
  hbox#pack ~expand:true ~fill:true vbox#coerce;

  let title = "<b><big>" ^ title ^ "</big></b>" in
  let label_title = GMisc.label ~yalign:0. ~line_wrap:true ~xalign:0. ~selectable:false ~markup:title
    ~packing:vbox#pack () in
  label_title#misc#set_can_focus false ;

  let label_info = GMisc.label ~yalign:0. ~line_wrap:true ~xalign:0. ~selectable:false ~markup:(info ^ "\n")
    ~packing:vbox#pack () in

  label_info#misc#set_can_focus false ;

  let frame = GBin.frame ~border_width:3 ~shadow_type:`IN ~packing:vbox#add () in
  let s = GBin.scrolled_window ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:frame#add  () in
  let () =
    let v = GTree.view ~packing:s#add
      ~fixed_height_mode:true
      ~reorderable:false
      ~rules_hint:true
      ~headers_visible:false
      ~enable_search:false
      () in
    ignore( v#append_column view_info );
    let store =  GTree.tree_store cols in
    fill_store store tree;
    v#misc#set_can_focus false;
    v#selection#set_mode `NONE;
    v#set_model (Some store#coerce) ;
    v#expand_all ();
  in
  s#misc#set_size_request ~height:(height - 20) ~width:(width - 40) () ;

  if is_error = false then
    w#add_button_stock `CANCEL `CANCEL;

  w#add_button_stock `OK `OK;

  let ret = ref false in
  begin match w#run () with
  | `OK -> ret:=true ; w#destroy ()
  | _ -> w#destroy ()
  end;
  !ret

let show_error ?width ?height ~tree ~window_title ~title ~info () =
  let _x = run ~is_error:true ?width ?height ~tree ~window_title ~title ~info () in
  ()

let ask ?width ?height ~tree ~window_title ~title ~info () =
  run ~is_error:false ?width ?height ~tree ~window_title ~title ~info ()

(*
let () =
  let tree = [
    ("1",true,["a";"b";"c"]);
    ("2",false,["a";"b";"c"]);
    ("3",true,["a";"b";"c"]);
    ("4",false,["a";"b";"c"]);
    ("5",false,["a";"b";"c"])
  ] in
  let _ = show_error ~tree ~window_title:"Window Title" ~title:"Apply changes" ~info:"Hahahaha?" () in
  ()
*)
