type t = {
  f_ok: unit -> unit ;
  f_err: unit -> unit ;
  job: Run.t;
}


let dummy_out (_s:string) = ()
let dummy_end () = ()



type mode =
| Sync
| Async
| Same_thread

let prepare ?(mode=Async) ?env ?(f_output=dummy_out)
    ?(f_finish_ok=dummy_end) ?(f_finish_err=dummy_end) ~command ~args () =

  let do_read = `Fun (match mode with
  | Async -> GtkThread.async f_output
  | Sync -> GtkThread.sync f_output
  | Same_thread -> f_output)
  in
  let stdout = do_read
  and stderr = do_read
  and args = Array.to_list args
  in
  let job = Run.new_command ?env ~stdin:`Null ~stdout ~stderr command args in
  {
    f_ok = f_finish_ok;
    f_err = f_finish_err;
    job
  }


let run (jo:t) =
  let run () =
    Run.run jo.job;
    match Run.close jo.job with
    | 0 -> GtkThread.async jo.f_ok ()
    | _ ->  GtkThread.async jo.f_err ()
  in
  ignore ( Thread.create run () )


let stop (jo:t) =
  Run.kill jo.job
