open Pipe

external exec_name : unit -> string = "exec_name"

let executable_name = match Sys.os_type with
| "Win32" -> exec_name ()
| _ -> Sys.executable_name

type split =
  | Nosplit
  | First_last of string * string


let split_one str char =
  try
    let i = String.index str char in
    let key = String.sub str 0 i
    and value = String.sub str (succ i) (String.length str - i - 1 ) in
    First_last(key,value)
  with
  | Not_found -> Nosplit


let split_rev str chr =
  let len = String.length str
  and ret = ref []
  and i = ref 0 in
  try
    while !i < len do
      let npos = String.index_from str !i chr in
      let nelem = String.sub str !i (npos - !i) in
      ret := nelem :: !ret ;
      i:= succ npos ;
    done;
    !ret
  with
    | Not_found ->
      let last = String.sub str !i (len - !i) in
      last::!ret


let split str chr = List.rev ^$ split_rev str chr


let is_space = function
|' ' |'\t' | '\n' | '\x0B' | '\x0C' | '\r' -> true
| _ -> false


let contains_spaces str =
  let rec iter str pos =
    if pos < 0 then
      false
    else if is_space str.[pos] then
      true
    else
      iter str (pred pos)
  in
  iter str (String.length str - 1)

let array_contains ar needle =
  let rec iter ar needle pos =
    if pos < 0 then
      false
    else if ar.(pos) = needle then
      true
    else
      iter ar needle (pred pos)
  in
  iter ar needle (Array.length ar - 1 )


let remove_spaces_begin_end str =
  let rec iter_end str i =
    if i < 0 then
      0
    else if is_space ^$ String.unsafe_get str i then
      iter_end str (pred i)
    else
      succ i
  in
  let rec iter_start str i = (* be careful, don't call me on strings that only contains space *)
    if is_space ^$ String.get str i then
      iter_start str (succ i)
    else
      i
  in
  let olen = String.length str in
  let nlen = iter_end str (pred olen) in
  if nlen = 0 then
    ""
  else
    let str_start = iter_start str 0 in
    if str_start = 0 && olen = nlen then
      str
    else
      String.sub str str_start (nlen-str_start)





let rec add_unique htl acc = function
  | [] -> List.rev acc
  | hd :: tl when not (Hashtbl.mem htl hd) ->
      Hashtbl.add htl hd ();
      add_unique htl (hd :: acc) tl
  | _ :: tl ->
      add_unique htl acc tl

let list_unique l =
  let h = Hashtbl.create ^$ List.length l in
  add_unique h [] l



let rec iter accu n = function
| [] -> List.rev accu, []
| l when n <= 0 -> List.rev accu,l
| hd::tl -> iter (hd::accu) (pred n) tl

let list_partition_by_number l n = iter [] n l



let try_finally f p g o =
  let module T = struct
    type 'a t =
    | Ok of 'a
    | Err of exn
  end in
  let erg = try T.Ok ( f p ) with x -> T.Err x in
  g o;
  match erg with
  | T.Ok x -> x
  | T.Err x -> raise x

let array_of_rev_list l =
  let ar = Array.of_list l in
  let n = Array.length ar in
  let j = ref (n-1) in
  for i = 0 to n/2 -1 do
    let x = ar.(i) in
    ar.(i) <- ar.(!j);
    ar.(!j) <- x;
    decr j
  done ;
  ar

let array_exists p ar =
  let rec iter p ar n i =
    if i = n then false
    else if p ar.(i) then true
    else iter p ar n (succ i)
  in
  iter p ar (Array.length ar) 0


let is_windows =   match Sys.os_type with
| "Win32" -> true
| _ -> false


let windows_mixed_path str = match is_windows with
| false -> str
| true ->
  let rec iter i str =
    if i >= 0 then (
      (match str.[i] with
      | '\\'  -> str.[i] <- '/'
      | _  -> ());
      iter (pred i) str
    )
  in
  let str = String.copy str in
  iter (String.length str - 1) str ;
  str



(*let replaced_environment ~name ~value =
  let found = ref false
  and env = Unix.environment ()
  and name_len = String.length name
  and accu = ref []  in

  for i = Array.length env - 1 downto 0 do
    let cur = env.(i) in
    let cur_len = String.length cur
    and replace = ref false  in
    if cur_len > name_len && cur.[name_len] = '=' then (
      let j = ref (pred name_len )
      and further = ref true in
      while !j >= 0 && !further = true do
        if cur.[!j] != name.[!j] then
          further:=false
        else
          decr j
      done;
      replace := !j = (-1);
    );
    if !replace then (
      accu := (name ^ "=" ^ value)::!accu ;
      found := true
    )
    else
      accu := cur::!accu
  done;
  if !found = false then
    accu:= (name ^ "=" ^ value)::!accu ;
  Array.of_list !accu
*)

let environment_with_removed_values names =
  let names = Array.of_list names in

  let env = Unix.environment ()
  and names_len = Array.map String.length names
  and n_names = Array.length names
  and accu = ref []  in

  for i = Array.length env - 1 downto 0 do
    let cur = env.(i) in
    let cur_len = String.length cur
    and remove = ref false  in

    for j = n_names - 1 downto 0 do
      let name_len = names_len.(j) in
      if cur_len > name_len && cur.[name_len] = '=' then
        let j = ref (pred name_len )
        and further = ref true
        and name = names.(j) in
        while !j >= 0 && !further = true do
          if cur.[!j] != name.[!j] then
            further:=false
          else
            decr j
        done;
        if  !j = (-1) then
          remove:=true ;
    done;
    if !remove = false then (
      accu := cur::!accu
    )
  done;
  Array.of_list !accu
