open Common

let always_true (_:int) = true
let always_false (_:int) = false

module Source=struct
  (* type t = {mutable checked: bool; mutable lname: string; } *)

  let column_list = new GTree.column_list ;;

  (*let cols = new GTree.column_list *)
  (*let col_ident = column_list#add Gobject.Data.int*)
  let col_text_name = column_list#add Gobject.Data.string
  let col_text_version_installed = column_list#add Gobject.Data.string
  let col_text_version_newest = column_list#add Gobject.Data.string
  let col_text_description = column_list#add Gobject.Data.string
  let col_cell_background_set = column_list#add Gobject.Data.boolean
  let col_cell_background = column_list#add Gobject.Data.string
  (* let col_pixbuf =  column_list#add ^$ Gobject.Data.gobject_by_name "GdkPixbuf" *)
  (* GdkPixbuf.pixbuf GTree.column = column_list#add ^$ Gobject.Data.gobject_by_name "GdkPixbuf" *)


  (* class based on the according examples from lablgtk2 *)
  class list_class =
  object (self)
    inherit
      [int,int,unit,unit] GTree.custom_tree_model column_list

    method custom_encode_iter cr = cr, (), ()
    method custom_decode_iter cr () () = cr

    val mutable cnts = [| |]
    val mutable cnts_orig = [| |]

    val mutable sort_func = None
    val mutable filter_func = always_true


    val htl_cnts = Hashtbl.create 10

    method private find_opt i =
      let len = Array.length cnts in
      if len < 1 || len <= i then
        None
      else
        Some(i)

    method custom_flags = [`LIST_ONLY]
    method custom_get_iter (path:Gtk.tree_path) : int option =
      let indices: int array  = GTree.Path.get_indices path in
      match indices with
      | [||] ->
          None
      | [|i|] -> self#find_opt i
      | _ -> failwith "Invalid Path of depth > 1 in a list"

    method get_i (path:Gtk.tree_path) : int =
      match GTree.Path.get_indices path with
      | [|i|] ->
        let len = Array.length cnts in
        if len < 1 || len <= i then
          failwith "invalid path get_i" ;
        cnts.(i)
      | _ -> failwith "invalid path get_i"

    method custom_get_path (row:int) : Gtk.tree_path =
      GTree.Path.create [row]

    method custom_value (_t:Gobject.g_type) (row:int) ~column =
      let i = cnts.(row) in
      let cur = !Gl.pkgs.(i)
      and todo = !Gl.to_do.(i) in

      match column with
      | 0 -> `STRING(Some cur.name)
      | 1 ->
        let show = match cur.status with
        | Current_version x
        | Local_version x
        | Old_version x -> Some x.i_pretty_version
        | Unused -> None
        in
        `STRING(show)
      | 2 -> `STRING(Some cur.pretty_version)
      | 3 -> `STRING(Some cur.comment)
      | 4 -> `BOOL (todo <> Dont_touch)
      | 5 ->
        let color = match todo with
        | Dont_touch -> None
        | Depending_install  | Depending_upgrade  | Depending_reinstall -> Some Const.to_install_dependent_color
        | Install | Upgrade | Reinstall -> Some Const.to_install_color
        | Depending_remove  -> Some Const.to_remove_dependent_color
        | Remove -> Some Const.to_remove_color
        in
        `STRING(color)
(*      | 6 ->
        (*let module M = struct
          external get_pointer : 'a -> Gpointer.boxed = "ml_g_value_get_pointer"
        end in
          `POINTER(Obj.magic(pixbuf_package_available)); *)
        `POINTER(Some (Obj.magic(pixbuf_package_available))); *)
        (* `CAML (Obj.repr pixbuf_package_available) *)
      | _ -> assert false


    method custom_iter_next (row:int) : int option =
      self#find_opt ^$ succ row

    method inform i =
      let n = try Hashtbl.find htl_cnts i with Not_found -> -1 in
      (* (-1) handled by binary. Todo: inform binary! *)
      assert ( n < Array.length cnts );
      if n >= 0 then
        let path = self#custom_get_path n in
        self#custom_row_changed path n

    method custom_iter_children (rowopt:int option) :int option =
      match rowopt with
      | None -> self#find_opt 0
      | Some _ -> None

    method custom_iter_has_child (_:int) : bool = false

    method custom_iter_n_children (rowopt:int option) : int =
      match rowopt with
      | None -> Array.length cnts
      | Some _ -> assert false

    method custom_iter_nth_child (rowopt:int option) (n:int)  : int option =
      match rowopt with
      | None -> self#find_opt n
      | _ -> None

    method custom_iter_parent (_:int) : int option = None

    method private update_htl_cnts () =
      Hashtbl.clear htl_cnts ;
      for i = 0 to Array.length cnts - 1 do
        Hashtbl.replace htl_cnts cnts.(i) i
      done

    method set_data () =
      let cts = ref [] in
      let ar = !Gl.pkgs in
      for i= Array.length ar - 1 downto 0 do
        let cur = ar.(i) in
        match cur.package_type with
        | PSource _ ->
          cts := i::!cts;
        | PBinary _ -> ()
      done;
      let ar = Array.of_list !cts in
      cnts <- ar;
      cnts_orig <- ar;
      self#resort_and_filter ()


    method private resort_and_filter () =
      let n_cnts = Array.copy cnts_orig
      and o_cnts = Array.copy cnts in

      let n_cnts =
        if filter_func == always_true then
          n_cnts
        else
          let n = ref [] in
          for i = Array.length n_cnts - 1 downto 0 do
            let j = n_cnts.(i) in
            if filter_func j then
              n := j ::!n ;
          done;
          Array.of_list !n
      in
      ( match sort_func with
      | None -> ()
      | Some f -> Array.sort f n_cnts );

      cnts <- n_cnts ;
      self#update_htl_cnts ();

      let len_o_cnts = Array.length o_cnts - 1
      and len_n_cnts = Array.length n_cnts - 1 in

      let max_len = max len_n_cnts len_o_cnts
      and i = ref 0 in
      while !i < max_len do
        let j = !i in
        incr i;
        let path = GTree.Path.create [j] in
        if j > len_n_cnts then (
          self#custom_row_deleted path
        )
        else if j > len_o_cnts then (
          self#custom_row_inserted path j
        )
        else
          let new' = n_cnts.(j)
          and old = o_cnts.(j) in
          if old <> new' then (
            self#custom_row_changed path j;
          )
      done

    method set_sort_func f =
      sort_func <- f;
      self#resort_and_filter ()

    method set_filter_func f =
      filter_func <- f;
      self#resort_and_filter ()

    method clear () =
      let ocnts = cnts in
      cnts_orig <- [| |];
      cnts <- [| |];
      Hashtbl.clear htl_cnts ;
      for i = 0 to Array.length ocnts - 1 do
        self#custom_row_deleted ^$ GTree.Path.create [i]
      done

  end


  let slist = new list_class
end


module Binary = struct

  let column_list = new GTree.column_list
  let col_name = column_list#add Gobject.Data.string
  let col_version = column_list#add Gobject.Data.string
  let col_comment = column_list#add Gobject.Data.string
  let col_cell_background_set = column_list#add Gobject.Data.boolean
  let col_cell_background = column_list#add Gobject.Data.string
  let col_pixbuf_visible = column_list#add Gobject.Data.boolean

  type tree =  {
    finfo: int; (* position in Gl.pkgs *)
    text: string option; (* what to display in the name column *)
    fidx: int; (* invariant: parent.(fidx)==myself *)
    parent: tree option;
    mutable sons: tree array;
  }

  type category = string * name list
  and name = string * version list
  and version = string * build list
  and build = int

  (* again based on example code from lablgtk2 *)
  class my_tree =
    let inbound i a = i>= 0 && i < Array.length a
    and htl_i = Hashtbl.create 10 in
  object (self)
    inherit
      [tree,tree,unit,unit] GTree.custom_tree_model column_list

    val mutable roots :  tree array = [||]
    val mutable unfiltered = [ ]
    val mutable filter_func = always_true

    method custom_encode_iter cr = cr, (), ()
    method custom_decode_iter cr () () = cr

    method custom_flags : GtkEnums.tree_model_flags list = [`ITERS_PERSIST]

    method custom_get_iter (path:Gtk.tree_path) : tree option =
      let indices: int array  = GTree.Path.get_indices path in
      match indices with
      | [||] -> None
      | _ ->
        if inbound indices.(0) roots = false then
          None
        else
          let result = ref (roots.(indices.(0))) in
          try
            for depth=1 to Array.length indices - 1 do
              let index = indices.(depth) in
              if inbound index !result.sons then
                result:=!result.sons.(index)
              else
                raise Not_found
            done;
            Some !result
          with
          | Not_found -> None

    method get_i (path:Gtk.tree_path) : int =
      match self#custom_get_iter path with
      | None ->  failwith "invalid path get_i"
      | Some i -> i.finfo

    method custom_get_path (row:tree) : Gtk.tree_path =
      let current_row = ref row in
      let path = ref [] in
      while !current_row.parent <> None do
        path := !current_row.fidx::!path;
        current_row := match !current_row.parent with Some p -> p
        | None -> assert false
      done;
      GTree.Path.create ((!current_row.fidx)::!path)

    method inform i =
      try
        let n = Hashtbl.find htl_i i in
        let path = self#custom_get_path n in
        self#custom_row_changed path n
      with
      | Not_found ->  () (* handled by source. Todo: inform binary! *)

    method custom_value (_t:Gobject.g_type) (row:tree) ~column =
      match column with
      | 0 ->
        `STRING(row.text)
      | 1 ->
        let show = match row.finfo with
        | (-1) -> None
        | i -> Some !Gl.pkgs.(i).pretty_version
        in
        `STRING(show)
      | 2 ->
        let show = match row.finfo with
        | (-1) -> None
        | i -> Some !Gl.pkgs.(i).comment
        in
        `STRING(show)
      | 3 ->
        let show = match row.finfo with
        | (-1) -> false
        | i -> !Gl.to_do.(i) <> Dont_touch
        in
        `BOOL (show)
      | 4 ->
        let color = match row.finfo with
        | (-1) -> None
        |  i ->
          match !Gl.to_do.(i) with
          | Dont_touch -> None
          | Depending_install  | Depending_upgrade  | Depending_reinstall -> Some Const.to_install_dependent_color
          | Install | Upgrade | Reinstall -> Some Const.to_install_color
          | Depending_remove  -> Some Const.to_remove_dependent_color
          | Remove -> Some Const.to_remove_color
        in
        `STRING(color)
      | 5 ->
        let show = match row.finfo with
        | (-1) ->  false
        | _ -> true
        in
        `BOOL (show)
      | _ -> assert false

    method custom_iter_next (row:tree) : tree option =
      let nidx = succ row.fidx in
      match row.parent with
      | None -> if inbound nidx roots then Some roots.(nidx)
        else None
      | Some parent ->
        if inbound nidx parent.sons then
          Some parent.sons.(nidx)
        else None

    method custom_iter_children (rowopt:tree option) :tree option =
      match rowopt with
      | None -> if inbound 0 roots then Some roots.(0) else None
      | Some row -> if inbound 0 row.sons then Some row.sons.(0) else None

    method custom_iter_has_child (row:tree) : bool =
      Array.length row.sons  > 0

    method custom_iter_n_children (rowopt:tree option) : int =
      match rowopt with
      | None -> Array.length roots
      | Some row -> Array.length row.sons

    method custom_iter_nth_child (rowopt:tree option) (n:int)
      : tree option =
      match rowopt with
      | None when inbound n roots -> Some roots.(n)
      | Some row when inbound n row.sons -> Some (row.sons.(n))
      | _ -> None

    method custom_iter_parent (row:tree) : tree option =
      row.parent

    method set_filter_func f =
      filter_func <- f;
      self#set_data unfiltered

    method set_data (tr:category list) =
      for i = Array.length roots - 1 downto 0 do
        let path = self#custom_get_path roots.(i) in
        self#custom_row_deleted path
      done;
      Hashtbl.clear htl_i ;

      unfiltered <- tr;

      let convert_tree tr =

        let rec get_builds parent counter accu = function
        | [] ->  Util.array_of_rev_list accu
        | i::tl ->
          if filter_func != always_true &&
            false = filter_func i
          then
            get_builds parent counter accu tl
          else
            let e = {
              finfo = i;
              text = None;
              parent = Some parent ;
              fidx = counter ;
              sons = [| |];
            }
            in
            Hashtbl.replace htl_i i e ;
            get_builds parent (succ counter) (e::accu) tl
        in

        let rec get_versions parent counter accu = function
        | [] ->  Util.array_of_rev_list accu
        | ((c,cl):version)::tl ->
          let e = {
            finfo = -1 ;
            text = Some c ;
            parent = Some parent ;
            fidx = counter ;
            sons = [| |];
          }
          in
          match get_builds e 0 [] cl with
          | [| |] -> get_versions parent counter accu tl
          | sl ->
            let e = match sl with
            | [|e'|] ->
              let i = e'.finfo in
              let e = { e with finfo = i } in
              Hashtbl.replace htl_i i e ;
              e
            | _ -> e.sons <- sl; e
            in
            get_versions parent (succ counter) (e::accu) tl
        in

        let rec get_names parent counter accu = function
        | [] ->  Util.array_of_rev_list accu
        | (c,cl)::tl ->
          let e = {
            finfo = -1 ;
            text = Some c ;
            parent = Some parent ;
            fidx = counter ;
            sons = [| |];
          }
          in
          match get_versions e 0 [] cl with
          | [| |] -> get_names parent counter accu tl
          | sl ->
            let e = match sl with
            | [| e' |] when Array.length e'.sons < 2 ->
              let i = e'.finfo in
              let e = { e with finfo = i } in
              Hashtbl.replace htl_i i e ;
              e
            | _ -> e.sons <- sl; e
            in
            get_names parent (succ counter) (e::accu) tl
        in
        let rec get_categories counter accu = function
        | [] -> Util.array_of_rev_list accu
        | (c,cl)::tl ->
          let nelem = {
            finfo = (-1) ;
            text = Some c;
            parent = None ;
            fidx = counter ;
            sons = [| |] ;
          }
          in
          match get_names nelem 0 [] cl with
          | [| |] ->
            get_categories counter accu tl
          | sl ->
            nelem.sons <- sl;
            get_categories (succ counter) (nelem::accu) tl
        in
        get_categories 0 [] tr
      in
      roots <- convert_tree tr

  end

  let get_htls () =
    let cat_htl = Hashtbl.create 5 in
    let pkgs = !Gl.pkgs in
    for i = 0 to Array.length pkgs - 1 do
      let cur = pkgs.(i) in
      match cur.package_type with
      | PSource _ -> ()
      | PBinary _ ->
        let name = cur.name in
        let category,name = match Util.split_one name '-' with
        | Util.Nosplit -> "uncategorized",name
        | Util.First_last(s1,s2) -> s1,s2
        in
        let name_htl =
          try
            Hashtbl.find cat_htl category
          with
          | Not_found -> let h = Hashtbl.create 3 in
                        Hashtbl.add cat_htl category h ;
                        h
        in
        let version_htl =
          try
            Hashtbl.find name_htl name
          with
          | Not_found -> let h = Hashtbl.create 3 in
                        Hashtbl.add name_htl name h ;
                        h
        in
        let version = cur.version in
        let builds =
          try
            Hashtbl.find version_htl version
          with
          | Not_found -> []
        in
        Hashtbl.replace version_htl version (i::builds) ;
    done;
    cat_htl

  let get_tree () =
    let compare_first (a,_) (b,_) = compare a b in
    let get_versions version l_builds accu =
      let cl = List.sort compare l_builds in
      (version,cl)::accu
    in
    let get_names name htl_versions accu =
      let compare_v (a,_) (b,_) = Ptype.compare_version a b
      and make_pretty (a,b) = Ptype.string_of_version a, b in
      let cl = List.map make_pretty ^$ List.sort compare_v ^$
        Hashtbl.fold get_versions htl_versions []
      in
      (name,cl)::accu
    in
    let get_cats category htl_names accu =
      let cl = List.sort compare_first ^$
        Hashtbl.fold get_names htl_names []
      in
      (category,cl)::accu
    in
    let cats = get_htls () in
    let l = List.sort compare_first ^$ Hashtbl.fold get_cats cats [] in
    (*let f (a,b) =
      let f (a,b) =
          let f (a,b) =
            let f i =
              Printf.printf "      %d\n" i;
            in
            print_string "    ";
            print_endline a;
            List.iter f b ;
          in
          print_string "  ";
          print_endline a;
          List.iter f b ;
      in
      print_endline a ;
      List.iter f b;
    in
    List.iter f l; *)
    l

  let slist = new my_tree
end



let get_pixbuf i =
  match !Gl.to_do.(i) with
  | Depending_install | Install -> Const.pixbuf_package_install
  | Depending_upgrade  |  Upgrade -> Const.pixbuf_package_upgrade
  | Depending_remove  |  Remove  -> Const.pixbuf_package_remove
  | Depending_reinstall | Reinstall -> Const.pixbuf_package_install
  | Dont_touch ->
    match !Gl.pkgs.(i).status with
    | Local_version _ | Current_version _ -> Const.pixbuf_package_installed_and_updated
    | Unused -> Const.pixbuf_package_available
    | Old_version _ -> Const.pixbuf_package_installed_outdated

let inform i =
  Binary.slist#inform i ;
  Source.slist#inform i




module Source_init = struct
  open Source
  let v = GTree.view ~fixed_height_mode:true ~reorderable:true ~rules_hint:true ~enable_search:false ()

  let view_is_installed = GTree.view_column ~title:"" ()
  let view_program_name = GTree.view_column ~title:"Program Name" ()

  let add_bg r (col:GTree.view_column) =
    col#add_attribute r "cell-background" col_cell_background ;
    col#add_attribute r "cell-background-set" col_cell_background_set

  let () =
    let r = GTree.cell_renderer_pixbuf [] in
    let cell_data_func (model:GTree.model) (row:Gtk.tree_iter) =
      let i = slist#get_i ^$ model#get_path row in
      let pixbuf = get_pixbuf i in
      r#set_properties [`PIXBUF pixbuf]
    in
    view_is_installed#pack r ;
    add_bg r view_is_installed;
    view_is_installed#set_sizing `FIXED;
    view_is_installed#set_cell_data_func r cell_data_func;
    view_is_installed#set_min_width 20;
    view_is_installed#set_clickable true;

    let f_ascend i j =
      let p2i = function
      | Install -> 0
      | Remove  -> 1
      | Upgrade -> 2
      | Reinstall -> 3
      | Depending_install -> 4
      | Depending_remove  -> 5
      | Depending_upgrade  -> 6
      | Depending_reinstall -> 7
      | Dont_touch -> 8
      and s2i = function
      | Old_version _ -> 0
      | Local_version _ -> 1
      | Current_version _ -> 2
      | Unused -> 3
      in
      let to_do = !Gl.to_do in
      let t1 = p2i to_do.(i)
      and t2 = p2i to_do.(j) in
      match compare t1 t2 with
      | 0 ->
        let pkgs = !Gl.pkgs in
        let t1 = pkgs.(i)
        and t2 = pkgs.(j) in
        let n1 = s2i t1.status
        and n2 = s2i t2.status in
        (match compare n1 n2 with
        | 0 -> compare t1.name t2.name
        | x -> x )
      | x -> x
    in
    let f_descend idx1 idx2 = f_ascend idx2 idx1 in

    let callback_header_clicked () =
      view_program_name#set_sort_indicator false ;
      let f =
        if view_is_installed#sort_indicator = false then (
          view_is_installed#set_sort_order `ASCENDING;
          view_is_installed#set_sort_indicator true;
          Some f_ascend
        )
        else (
          match view_is_installed#sort_order with
          | `ASCENDING ->
            view_is_installed#set_sort_order `DESCENDING ;
            view_is_installed#set_sort_indicator true;
            Some f_descend
          | `DESCENDING ->
            view_is_installed#set_sort_indicator false ;
            None
        )
      in
      slist#set_sort_func f
    in
    ignore ( view_is_installed#connect#clicked ~callback:callback_header_clicked ) ;

    let r = GTree.cell_renderer_text [] in
    view_program_name#pack r;
    view_program_name#add_attribute r "text" col_text_name;
    add_bg r view_program_name;
    view_program_name#set_sizing `FIXED;
    view_program_name#set_min_width 100;
    view_program_name#set_resizable true;
    view_program_name#set_reorderable true;
    view_program_name#set_clickable true;
    view_program_name#set_fixed_width 150 ;

    let f_ascend i j =
      let pkgs = !Gl.pkgs in
      let n1 = pkgs.(i).name
      and n2 = pkgs.(j).name in
      String.compare n1 n2
    in
    let f_descend i j = f_ascend j i in

    let callback () =
      view_is_installed#set_sort_indicator false ;
      let f =
        if view_program_name#sort_indicator = false then
          let () = view_program_name#set_sort_order `ASCENDING in
          view_program_name#set_sort_indicator true;
          Some f_ascend
        else
          match view_program_name#sort_order with
          | `ASCENDING ->
            view_program_name#set_sort_order `DESCENDING ;
            Some f_descend
          | `DESCENDING ->
            view_program_name#set_sort_indicator false ;
            None
      in
      slist#set_sort_func f
    in
    let _ = view_program_name#connect#clicked ~callback in
    ()

  let handle_change index new' =
    let module G = Ginfo in
    let to_do = !Gl.to_do in
    let new_to_do = Array.copy to_do
    and old_to_do = Array.copy to_do in
    begin
      let cur = !Gl.pkgs.(index) in
      match Ptype.get_binary_info cur with
      | None -> ()
      | Some x ->
        try
          let i = Hashtbl.find Gl.ident_bin x.b_ident in
          match new' with
          | Dont_touch
          | Remove -> new_to_do.(i) <- new'
          | Upgrade
          | Reinstall -> new_to_do.(i) <- Depending_remove
          | Install | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall
            -> assert false
        with
        | Not_found -> ()
    end;
    new_to_do.(index) <- new' ;
    if Ginfo.try_apply_to_do new_to_do then
      for idx = 0 to Array.length to_do - 1 do
        if index = idx || to_do.(idx) <> old_to_do.(idx) then
          inform idx
      done


  let () =
    v#selection#set_mode `SINGLE;
    v#misc#set_sensitive false ;

    let view_program_status  =
      let r = GTree.cell_renderer_text [] in
      let col = GTree.view_column ~title:"Installed version" ~renderer:(r,["text",col_text_version_installed]) () in
      add_bg r col  ;
      col#set_sizing `FIXED;
      col#set_min_width 20;
      col#set_resizable true;
      col#set_reorderable true;
      col#set_fixed_width 115;
      col

    and view_newest_version =
      let r = GTree.cell_renderer_text [] in
      let col = GTree.view_column ~title:"Newest Version" ()
        ~renderer:(r,["text",col_text_version_newest])
      in
      add_bg r col;
      col#set_sizing `FIXED;
      col#set_min_width 100;
      col#set_resizable true;
      col#set_reorderable true;
      col#set_fixed_width 115 ;
      col

    and view_description =
      let r = GTree.cell_renderer_text [] in
      let col = GTree.view_column ~title:"Description" ()
        ~renderer:(r,["text",col_text_description])
      in
      add_bg r col;
      col#set_sizing `FIXED;
      col#set_min_width 100;
      col#set_resizable true;
      col#set_reorderable true;
      col
    in

    ignore( v#append_column view_is_installed );
    ignore( v#append_column view_program_name );
    ignore( v#append_column view_program_status );
    ignore( v#append_column view_newest_version );
    ignore( v#append_column view_description );

    let _ = v#selection#connect#after#changed ~callback:(fun () ->
      match v#selection#get_selected_rows with
      | [] -> ()
      | [a ] ->
        Notebook.set ^$ slist#get_i a;
        ()
      | _ -> assert false
    )
    in

    let _ = v#event#connect#button_press ~callback:( fun ev ->
      if GdkEvent.Button.button ev <> 3 || GdkEvent.get_type ev <> `BUTTON_PRESS then
        false
      else
        let x = int_of_float ^$ GdkEvent.Button.x ev
        and y = int_of_float ^$ GdkEvent.Button.y ev in
        match v#get_path_at_pos ~x ~y with
        | None -> false
        | Some(path,_,_,_) ->
          let i = slist#get_i path in
          let menu = GMenu.menu ()
          and to_do = !Gl.to_do.(i)
          and pkg = !Gl.pkgs.(i) in
          let status = pkg.status in

          let um = GMenu.menu_item ~label:"Un_mark" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Install | Upgrade | Remove | Reinstall -> true
          | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall
          | Dont_touch -> false in

          if is_sensitive then
            ignore( um#connect#activate ~callback:(fun () -> handle_change  i Dont_touch ) )
          else
            um#misc#set_sensitive is_sensitive ;

          let ib = GMenu.menu_item ~label:"_Install" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Depending_reinstall   | Depending_remove  | Install -> false
          | Depending_install  | Depending_upgrade    (* Upgrade ?*)
          | Upgrade | Remove | Reinstall | Dont_touch ->
            match status with
            | Current_version _ | Local_version _ | Old_version _ -> false
            | Unused -> true
          in
          if is_sensitive then
            ignore( ib#connect#activate ~callback:(fun () -> handle_change  i Install ) )
          else
            ib#misc#set_sensitive is_sensitive ;

          let rib= GMenu.menu_item ~label:"R_e-Install" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Depending_install  | Depending_upgrade   | Depending_remove  | Depending_reinstall
          | Reinstall -> false
          | Upgrade | Remove | Install | Dont_touch ->
            match status with
            | Current_version _ -> true
            | Unused | Local_version _ | Old_version _ -> false
          in
          if is_sensitive then
            ignore( rib#connect#activate ~callback:(fun () -> handle_change  i Reinstall ) )
          else
            rib#misc#set_sensitive is_sensitive ;

          let ub= GMenu.menu_item ~label:"_Upgrade" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Depending_install  | Depending_reinstall  (* todo: handle these cases *)
          | Depending_upgrade   | Depending_remove
          | Upgrade -> false
          | Reinstall| Remove | Install | Dont_touch ->
            match status with
            | Old_version _  -> true
            | Current_version _ | Unused | Local_version _  -> false
          in
          if is_sensitive then
            ignore( ub#connect#activate ~callback:(fun () -> handle_change  i Upgrade ) )
          else
            ub#misc#set_sensitive is_sensitive ;

          let rb = GMenu.menu_item ~label:"_Remove" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Depending_install  | Depending_upgrade   | Depending_remove  | Depending_reinstall
          | Remove -> false
          | Reinstall| Upgrade | Install | Dont_touch ->
            match status with
            | Unused -> false
            | Current_version x | Old_version x  | Local_version x -> not x.i_preserve && pkg.name <> "godi-tools" && pkg.name <> "godi-tools-gui"
          in
          if is_sensitive then
            ignore( rb#connect#activate ~callback:(fun () -> handle_change i Remove ) )
          else
            rb#misc#set_sensitive is_sensitive ;

          menu#popup ~button:0 ~time:(GtkMain.Main.get_current_event_time ()) ;
          true
    ) in

    let _ = v#connect#after#row_activated ~callback:(fun path _vcol ->
      let index = slist#get_i path in
      let p = !Gl.pkgs.(index) in
      if p.status = Unused then (
        match !Gl.to_do.(index) with
        | Depending_install  | Depending_upgrade   | Depending_remove  | Depending_reinstall
        | Upgrade | Remove | Reinstall -> ()
        | (Install as old)| (Dont_touch as old) ->
           let new' = if old = Install then Dont_touch else Install in
           handle_change index new'
      )
    )
    in
    v#set_model (Some slist#coerce)


  class viewer =
  object(self)
    method misc = v#misc
    method coerce = v#coerce
    method drag = v#drag
    method as_widget = v#as_widget

    method set_filter_func f =
      v#set_model None;
      slist#set_filter_func f;
      v#set_model (Some slist#coerce);

    (* call me before (re-)fill the model *)
    method private clear () =
      v#set_model None ;
      slist#clear ()

    method deactivate () =
      self#clear () ;
      v#misc#set_sensitive false

    method activate () =
      self#clear () ;
      slist#set_data ();
      v#set_model (Some slist#coerce);
      if SHashtbl.length Gl.names_src > 0 then
        v#misc#set_sensitive true;
  end
end


module Binary_init = struct
  open Binary

  let add_bg r (col:GTree.view_column) =
    col#add_attribute r "cell-background" col_cell_background ;
    col#add_attribute r "cell-background-set" col_cell_background_set

  let pb_renderer = GTree.cell_renderer_pixbuf [`XALIGN 0. ; `YALIGN 0.5 ; `XPAD 0 ; `YPAD 0 ]
  let pb_cell_data_func (model:GTree.model) (row:Gtk.tree_iter) =
    match slist#get_i ^$ model#get_path row with
    | (-1) -> () (* properties := (`VISIBLE false)::!properties ; *)
    | i ->
      let pixbuf = get_pixbuf i in
      pb_renderer#set_properties [`PIXBUF pixbuf]

  let view_program_name =
    let col = GTree.view_column ~title:"Name" () in
    col#pack ~from:`START pb_renderer ;
    col#add_attribute pb_renderer "visible" col_pixbuf_visible;
    add_bg pb_renderer col;
    col#set_cell_data_func pb_renderer pb_cell_data_func;

    let renderer = GTree.cell_renderer_text [`XALIGN 0.; `YALIGN 0.5] in
    col#pack renderer ;
    col#add_attribute renderer "text" col_name;
    add_bg renderer col;

    col#set_sizing `FIXED;
    col#set_min_width 100;
    col#set_resizable true;
    col#set_reorderable true;
    col#set_fixed_width 160 ;
    col

  let view_program_version =
    let renderer = GTree.cell_renderer_text [] in
    let col = GTree.view_column ~title:"Version" () in
    col#pack renderer ;
    add_bg renderer col;
    col#add_attribute renderer "text" col_version;
    col#set_sizing `FIXED;
    col#set_min_width 100;
    col#set_resizable true;
    col#set_reorderable true;
    col#set_fixed_width 150 ;
    col

  let view_program_comment =
    let renderer = GTree.cell_renderer_text [] in
    let col = GTree.view_column ~title:"Comment" () in
    col#pack renderer ;
    add_bg renderer col;
    col#set_sizing `FIXED;
    col#set_min_width 100;
    col#set_resizable true;
    col#set_reorderable true;
    col#add_attribute renderer "text" col_comment;
    col

  let handle_change index new' =
    let to_do = !Gl.to_do in
    let new_to_do = Array.copy to_do
    and old_to_do = Array.copy to_do in
    if new' = Dont_touch || new' = Remove then begin
      try
        let cur = !Gl.pkgs.(index) in
        let i = SHashtbl.find Gl.names_src cur.name in
        if to_do.(i) = to_do.(index) then
          new_to_do.(i) <- new'
      with
      | Not_found -> ()
    end;
    new_to_do.(index) <- new' ;
    if Ginfo.try_apply_to_do new_to_do then
      let pkgs = !Gl.pkgs in
      for idx = 0 to Array.length pkgs - 1 do
        if index = idx || to_do.(idx) <> old_to_do.(idx) then
          inform idx
      done


  let v = GTree.view ~fixed_height_mode:true ~reorderable:true ~rules_hint:true ~enable_search:false ()

  let callback_selection_changed () =
    match v#selection#get_selected_rows with
    | [] -> ()
    | [a] ->
      (* -1: display no package *)
      Notebook.set ^$ slist#get_i a
    | _ -> assert false

  let callback_row_activated path _vcol =
    match slist#get_i path with
    | (-1) ->
      if v#row_expanded path then
        v#collapse_row path
      else
        v#expand_row path;
    | i ->
      let p = !Gl.pkgs.(i)
      and to_do = !Gl.to_do.(i) in
      if p.status = Unused then
        match to_do with
        | Depending_install  | Depending_upgrade   | Depending_remove  | Depending_reinstall
        | Upgrade | Remove | Reinstall -> ()
        | Install | Dont_touch ->
          let new' = if to_do = Install then Dont_touch else Install in
          handle_change i new'
      else
        match to_do with
        | Depending_install  | Depending_upgrade   | Depending_remove  | Depending_reinstall
        | Upgrade | Install | Reinstall -> ()
        | Remove  | Dont_touch ->
          let new' = if to_do = Remove then Dont_touch else Remove in
          handle_change i new'

  let callback_button_press ev =
    if GdkEvent.Button.button ev <> 3 || GdkEvent.get_type ev <> `BUTTON_PRESS then
      false
    else
      let x = int_of_float ^$ GdkEvent.Button.x ev
      and y = int_of_float ^$ GdkEvent.Button.y ev in
      match v#get_path_at_pos ~x ~y with
      | None -> false
      | Some(path,_,_,_) ->
        match slist#get_i path with
        | (-1) -> false
        | i ->
          let menu = GMenu.menu ()
          and to_do = !Gl.to_do.(i)
          and pkg = !Gl.pkgs.(i) in
          let status = pkg.status in

          let um = GMenu.menu_item ~label:"Un_mark" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Install | Upgrade | Remove | Reinstall -> true
          | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall
          | Dont_touch -> false in
          (match is_sensitive with
          | true -> ignore( um#connect#activate ~callback:(fun () -> handle_change  i Dont_touch ) )
          | false -> um#misc#set_sensitive is_sensitive );


          let ib = GMenu.menu_item ~label:"_Install" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Depending_reinstall   | Depending_remove  | Install -> false
          | Depending_install  | Depending_upgrade
          | Upgrade | Remove | Reinstall | Dont_touch ->
            match status with
            | Current_version _ | Local_version _ | Old_version _ -> false
            | Unused -> true
          in
          if is_sensitive then
            ignore( ib#connect#activate ~callback:(fun () -> handle_change  i Install ) )
          else
            ib#misc#set_sensitive is_sensitive ;

          let rb = GMenu.menu_item ~label:"_Remove" ~use_mnemonic:true ~packing:menu#append () in
          let is_sensitive = match to_do with
          | Depending_install  | Depending_upgrade   | Depending_remove  | Depending_reinstall
          | Remove -> false
          | Reinstall| Upgrade | Install | Dont_touch ->
            match status with
            | Unused -> false
            | Current_version x | Old_version x  | Local_version x -> not x.i_preserve && pkg.name <> "godi-tools" && pkg.name <> "godi-tools-gui"
          in
          if is_sensitive then
            ignore( rb#connect#activate ~callback:(fun () -> handle_change i Remove ) )
          else
            rb#misc#set_sensitive is_sensitive ;

          menu#popup ~button:0 ~time:(GtkMain.Main.get_current_event_time ()) ;
          true

  let () =
    v#selection#set_mode `SINGLE;
    v#misc#set_sensitive false ;

    ignore( v#append_column view_program_name );
    ignore( v#append_column view_program_version );
    ignore( v#append_column view_program_comment );

    ignore ( v#selection#connect#after#changed ~callback:callback_selection_changed ) ;
    ignore ( v#connect#after#row_activated ~callback:callback_row_activated ) ;
    ignore ( v#event#connect#button_press ~callback:callback_button_press );

    v#set_model (Some slist#coerce)



  let clear () =
    v#set_model None ;
    slist#set_data []

  class viewer =
  object
    method misc = v#misc
    method coerce = v#coerce
    method drag = v#drag
    method as_widget = v#as_widget

    method set_filter_func f =
      v#set_model None;
      slist#set_filter_func f;
      v#set_model (Some slist#coerce);
      v#expand_all ()

    method deactivate () =
      clear ();
      v#misc#set_sensitive false

    method activate () =
      clear () ;
      let tr = get_tree () in
      slist#set_data tr  ;

      v#set_model (Some slist#coerce);
      if SHashtbl.length Gl.names_bin > 0 then (
        v#misc#set_sensitive true;
      );
      (* why is this necessary? activate is already called via
         GtkThread.async! *)
      GtkThread.async v#expand_all ();
  end
end

type view = (<
              as_widget : Gtk.widget Gtk.obj ;
             coerce : GObj.widget  ;
             drag : GObj.drag_ops  ;
             misc : GObj.misc_ops  ;
             set_filter_func: (int -> bool) -> unit;
             activate: unit -> unit ;
             deactivate: unit -> unit
                         >)

let s_view = new Source_init.viewer
let s_view = ( s_view :>  view )

let b_view = new Binary_init.viewer
let b_view = ( b_view :>  view )
