open Common

let tb =
  let t = GButton.toolbar () in
  t#misc#set_can_focus false ;
  t

let apply_button =
  let t = GButton.tool_button ~stock:`APPLY ~packing:tb#insert () in
  t#misc#set_sensitive false ;
  t

let cancel_button =
  let t = GButton.tool_button ~stock:`CANCEL ~packing:tb#insert () in
  t#misc#set_sensitive false ;
  t

let update_button =
  let t = GButton.tool_button ~label:"Update" ~packing:tb#insert () in
  let image = GMisc.image ~pixbuf:Const.pixbuf_update () in
  t#set_icon_widget image#coerce ;
  t#misc#set_can_focus false ;
  t#misc#set_sensitive false ;
  t

let select_newer_button =
  let t = GButton.tool_button ~label:"Select newer" ~packing:tb#insert () in
  let image = GMisc.image ~pixbuf:Const.pixbuf_select_newer () in
  t#set_icon_widget image#coerce ;
  t#misc#set_can_focus false ;
  t#misc#set_sensitive false ;
  if Const.dont_show_source_packages then
    t#misc#hide ();
  t

let combo_source =
  let _ = GButton.separator_tool_item ~packing:tb#insert () in
  let dummy = GButton.tool_item ~packing:tb#insert () in
  let box = GPack.vbox ~spacing:5 ~homogeneous:true  ~packing:dummy#add  () in
  let strings = ""::Const.status_cat_updatable::Const.status_cat_installed::
    Const.status_cat_not_installed::(List.map snd Const.categories)
  in
  let (c,_) = GEdit.combo_box_entry_text ~strings ()  in
  box#pack ~expand:false ~fill:false c#coerce ;
  c#set_active 0 ;
  c#misc#hide ();
  c#misc#set_can_focus false ;
  c#misc#set_sensitive false;
  c#misc#set_size_request ~height:20 ();
  c

let combo_bin =
  let dummy = GButton.tool_item ~packing:tb#insert () in
  let _ = GButton.separator_tool_item ~packing:tb#insert () in
  let box = GPack.vbox ~spacing:5 ~homogeneous:true  ~packing:dummy#add  () in
  let strings = ""::Const.status_cat_installed::Const.status_cat_not_installed::[] in
  let (c,_) = GEdit.combo_box_entry_text ~strings ()  in
  box#pack ~expand:false ~fill:false c#coerce ;
  c#misc#hide ();
  c#set_active 0 ;
  c#misc#set_can_focus false ;
  c#misc#set_sensitive false;
  c#misc#set_size_request ~height:20 ();
  c

let license = "Godi: Copyright (c) 2003 Gerd Stolpmann\n\
\n\
This software is provided 'as-is', without any express or implied\n\
warranty. In no event will the authors be held liable for any damages\n\
arising from the use of this software.\n\
\n\
Permission is granted to anyone to use this software for any purpose,\n\
including commercial applications, and to alter it and redistribute it\n\
freely, subject to the following restrictions:\n\
\n\
1. The origin of this software must not be misrepresented; you must\n\
not claim that you wrote the original software. If you use this\n\
software in a product, an acknowledgment in the product documentation\n\
would be appreciated but is not required.\n\
\n\
2. Altered source versions must be plainly marked as such, and must\n\
not be misrepresented as being the original software.\n\
\n\
3. This notice may not be removed or altered from any source\n\
distribution.\n\
\n\
---\n\
\n\
godi-tools contains various third-party software:\n\
\n\
 - zlib (zlib License)\n\
 - camlzip (GNU Lesser General Public License with ocaml linking exception \n\
 - SHA1 by Steve Reid (public domain) \n\
\n\
\n\
The gtk-based GUI is an external contribution (CDDL).\n\
\n\
"


let about =
  let _ = GButton.separator_tool_item ~draw:false ~expand:true ~packing:tb#insert () in
  let t = GButton.tool_button ~stock:`ABOUT ~packing:tb#insert () in
  t#misc#set_can_focus false ;
  t#misc#set_sensitive false ;

  let run_dialog () =
    let dialog =
      GWindow.about_dialog
        ~name:"Godi "
        ~copyright:"Copyright: Gerd Stolpmann and others"
        ~license
        ~website:"https://github.com/arirux/godigui"
        ~website_label:"Github Repository GUI"
        ()
    in
    dialog#set_icon (Some Const.pixbuf_camel);
    ignore (dialog#connect#response ~callback:(fun _ -> dialog#destroy () ));
    try
      ignore ( dialog#run () );
      ()
    with
    | _ -> ()
  in
  let _ = t#connect#clicked ~callback:run_dialog in
  t
